<?php

namespace Drupal\inotherwords\Plugin\Field\FieldFormatter;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceLabelFormatter;
use Drupal\Core\Form\FormStateInterface;

/**
 * Plugin implementation of the 'inotherwords sequential terms' formatter.
 *
 * It applies to entity reference taxonomy terms.
 *
 * We're only set up to handle taxonomy terms right now, because we get the
 * full order from their vocabulary, but the same thing could be done for text
 * lists.  It would take a bit of a lift but ideally the comparison of one list
 * to another, after the items have been normalized, can be moved to a separate
 * PHP library.
 *
 * In any case, it's cleaner to make this work for text lists in a separate
 * formatter.
 *
 * @FieldFormatter(
 *   id = "inotherwords_sequential_terms",
 *   label = @Translation("In other words: Sequential terms"),
 *   description = @Translation("Summarize sequential taxonomy terms; given
 *   <em>Monday, Tuesday, Wednesday</em> return <em>Monday through Wednesday</em>."),
 *   field_types = {
 *     "entity_reference"
 *   }
 * )
 */
class InotherwordsSequentialTermsFormatter extends EntityReferenceLabelFormatter {

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'connecting_word' => 'through',
      'minimum_sequence' => 4,
      'text_before' => '',
      'text_after' => '',
      'all_items_text' => 'All @field_name',
      'oxford_comma' => TRUE,
      'link' => FALSE,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $elements['connecting_word'] = [
      '#title' => t('Connecting word'),
      '#description' => t('Word or symbol (such as "to" or "–") to replace the removed items in a sequence'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('connecting_word'),
    ];
    $elements['minimum_sequence'] = [
      '#title' => t('Minimum number of items in a sequence before removing middle items'),
      '#description' => t('Minimum number of items in a sequence before removing middle items'),
      '#type' => 'number',
      '#min' => 3,
      '#default_value' => $this->getSetting('minimum_sequence'),
    ];
    $elements['text_before'] = [
      '#title' => t('Text before sequence summary'),
      '#description' => ('Before all other output, start with this, for example: "Ages". <em>Not</em> added before the All items text.'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('text_before'),
    ];
    $elements['text_after'] = [
      '#title' => t('Text after sequence summary'),
      '#description' => ('After all other output end with this, for example: "years old". <em>Not</em> added after the All items text.'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('text_after'),
    ];
    $elements['oxford_comma'] = [
      '#title' => t('Connect sequences with natural language'),
      '#description' => ('Items within sequences that fall below the minimum number of items, and the sequences themselves, will be connected with commas, the oxford comma, and "and".  For example: "One to three, five and six, and eight".'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('oxford_comma'),
    ];
    $elements['all_items_text'] = [
      '#title' => t('All items text'),
      '#description' => t('Words to <em>replace</em> output with if every item was selected.  Make blank to not do this.'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('all_items_text'),
    ];
    $elements['link'] = [
      '#title' => t('Link label to the referenced entity'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('link'),
    ];

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];
    $summary[] = t('Series of at least @minimum_sequence items drop intermediary items in favor of "@connecting_word" as a connector',
      [
        '@minimum_sequence' => $this->getSetting('minimum_sequence'),
        '@connecting_word' => $this->getSetting('connecting_word'),
      ]
    );
    $all_items_text = $this->getSetting('all_items_text');
    $summary[] = $all_items_text ? t('If all items are selected the text output will be "@all_items_text"',
      ['@all_items_text' => $all_items_text]) : '';
    $summary[] = $this->getSetting('link') ? t('Try to link to referenced entities') : t('No link');
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = parent::viewElements($items, $langcode);
    // This will be what we return instead of elements.
    $inotherwords = [];

    // Not sure if we'd ever be given zero elements, but if so, give 'em back.
    if (!$elements) {
      return $elements;
    }

    $single = count($elements) === 1;

    $output_as_link = $this->getSetting('link');
    $connecting_word = $this->getSetting('connecting_word');
    $text_before = $this->getSetting('text_before');
    $text_after = $this->getSetting('text_after');
    $minimum_sequence = $this->getSetting('minimum_sequence');
    $all_items_text = $this->getSetting('all_items_text');

    $sequences = $this->getSequencedElements($elements, $items, $output_as_link, $langcode);

    if (!isset($sequences['all_items'])) {
      // We didn't actually make sequences, what was returned is just elements.
      return $sequences; // Equal to $elements.
    }
    $all_items = $sequences['all_items'];
    // Get sequences back to being a nested, but all-numeric-key, array.
    unset($sequences['all_items']);

    $handler_settings = $this->fieldDefinition->getSetting('handler_settings');
    $vocabulary_id = current($handler_settings['target_bundles']);

    if ($all_items && $all_items_text) {
      $field_name = $this->fieldDefinition->getLabel();
      $field_machine_name = $this->fieldDefinition->getFieldStorageDefinition()->getSetting('label');
      $inotherwords[0] = [
        '#markup' => new FormattableMarkup($all_items_text, [
          '@field_name' => $field_name,
          '@field_machine_name' => $field_machine_name,
          '@vocabulary_machine_name' => $vocabulary_id,
        ]),
      ];
      // @TODO if $output_as_link is TRUE make a hover-over with all the links.
      return $inotherwords;
    }

    if ($text_before) {
      $texts = explode('|', $text_before);
      if (count($texts) > 1) {
        $text_before = $single ? $texts[1] : $texts[0];
      }
      // Quotation marks around variable are intentional to add trailing space.
      $inotherwords[] = ['#markup' => "$text_before "];
    }

    foreach ($sequences as $sequence_id => $sequence) {
      if (count($sequence) >= $minimum_sequence) {
        $last_element = end($sequence);
        $first_element = reset($sequence);
        $inotherwords[$sequence_id][] = $first_element;
        $inotherwords[$sequence_id][] = ['#markup' => " $connecting_word "];
        $inotherwords[$sequence_id][] = $last_element;
      }
      else {
        // Each of the single items not in a consecutive sequence still need to
        // be output!
        $inotherwords[$sequence_id] = $sequence;
      }
    }

    if ($this->getSetting('oxford_comma')) {
      // Make every series, and the groupings of the series, grammatical.
      // Translate 'and' and 'or'.
      if ($connecting_word == 'and') {
        $type = t('and', [], ['context' => 'Final join']);
      }
      elseif ($connecting_word == 'or') {
        $type = t('or', [], ['context' => 'Final join']);
      }

    }

    if ($text_after) {
      $texts = explode('|', $text_after);
      if (count($texts) > 1) {
        $text_after = $single ? $texts[1] : $texts[0];
      }
      // Quotation marks around variable are intentional to add leading space.
      $inotherwords[] = ['#markup' => " $text_after"];
    }

    return $inotherwords;
  }

  /**
   * Returns elements sorted into sequences of consecutive items.
   *
   * @param \Drupal\Core\Entity\EntityInterface[]
   *   The array of referenced entities to display, keyed by delta.
   * @param \Drupal\Core\Field\EntityReferenceFieldItemListInterface $items
   *   The item list.
   * @param string $langcode
   *   The language code of the referenced entities to display.
   *
   * @return array $sequences
   *   The elements to display, nested in consecutive sequences.
   *
   * @see ::viewElements()
   */
  public function getSequencedElements($elements, FieldItemListInterface $items, $output_as_link, $langcode) {
    $handler_settings = $this->fieldDefinition->getSetting('handler_settings');
    $vocabulary_id = current($handler_settings['target_bundles']);
    $TermStorage = \Drupal::entityTypeManager()->getStorage('taxonomy_term');
    $ordered_terms = $TermStorage->loadTree($vocabulary_id);

    $label_key = ($output_as_link) ? '#title' : '#plain_text';
    $label = $elements[0][$label_key];
    // Get the location of the first item we're given to display in the full
    // source it was drawn from.  As of PHP, we can use the `use` keyword to
    // pass a value into the anonymous function that yields a closure.
    $first = array_filter($ordered_terms, function($term) use ($label) {
      return $term->name === $label;
    });
    $next_key = key($first);

    // If our first item isn't in the source, there's been some mistake somehow
    // so leave without doing anything and know if something blows up it's not
    // our fault.
    if ($next_key === NULL) {
      return $elements;
    }

    // We count on the given items and the full list of items to be in the same
    // order.  We'd have to re-order the given terms if that were not the case.

    $sequences = [];
    $sequence_id = 0;
    // By default figure every element from the full potential list is present.
    // And yes we're mixing associative array keys with numeric keys, but we'll
    // clean it up later.
    $sequences['all_items'] = TRUE;
    while (TRUE) {
      // Advance to our first or next matching item in the ordered items.
      if (key($ordered_terms) === $next_key) {
        // Add any items in sync with ordered items.
        $current_element = current($elements);
        // When we're out of elements to sequence, we can just return.
        if ($current_element === FALSE) {
          return $sequences;
        }

        $next_element = next($elements);
        if (current($ordered_terms)->name !== $current_element[$label_key]) {
          \Drupal::logger('inotherwords')->notice("Unexpected result that all terms term %name did not equal current element %title.", [
            '%name' => current($ordered_terms)->name,
            '%title' => $current_element[$label_key],
          ]);
        } else {
          // All is as expected, add the element to the current sequence.
          $sequences[$sequence_id][] = $current_element;
          // Use our next element's name to get matching ordered items' key.
          $label = $next_element[$label_key];
          $next = array_filter($ordered_terms, function($term) use ($label) {
            return $term->name === $label;
          });
          $next_key = key($next);
        }
      }
      else {
        // If we got here, we had to skip one of the ordered items.
        $sequence_id++;
        $sequences['all_items'] = FALSE;
      }
      // Remove any possibility of an infinite loop (but should never happen).
      if (key($ordered_terms) === NULL) {
        break;
      };
      next($ordered_terms);
    }

    return $sequences;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    // This formatter is only available for taxonomy terms.
    return $field_definition->getFieldStorageDefinition()->getSetting('target_type') == 'taxonomy_term';
  }

}

/**
 * Make an array of items into a proper, punctuated, and sentence-ready list.
 *
 * Ultimately we want this to live at https://gitlab.com/agaric/php/oxford-comma
 * but will workshop it here.
 *
 * @param $list
 *   An array of words or items to join.
 * @param $settings
 *   An array of optional settings to use in making the Oxford comma list:
 *   - type
 *     The text to use between the last two items. Defaults to 'and'. Pass in
 *     'or' and 'and' without translation; translate any other join.
 *   - comma
 *     The join for the list. Defaults to a comma followed by a space.
 *     To make an Oxford comma list with semicolons, use '; '.
 *   - oxford
 *     Change this from default 'TRUE' and you are a philistine.
 */
function oxford_comma_list($list, $settings = []) {
  // Set default settings.
  $comma = ', ';
  $type = 'and';
  $oxford = TRUE;
  // Overwrite default settings with any passed-in settings that apply.
  extract($settings, EXTR_IF_EXISTS);

  // If the array of items passed in has only one item in it, return that item.
  if (count($list) === 1) {
    return current($list);
  }

  if ($oxford && count($list) > 2) {
    $final_join = $comma . $type . ' ';
  }
  else {
    $final_join = ' ' . $type . ' ';
  }
  // Take the last two elements off of the $list array.
  $final = array_splice($list, -2, 2);
  // Combine the final two removed elements around the final join string.
  $final_string = implode($final_join, $final);
  // Add the combined elements (now a single element) back onto the list array.
  array_push($list, $final_string);
  // Return the list as a text string joined together with commas (or other).
  return implode($comma, $list);
}
